(*
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2015-2017)
         René Thiemann <rene.thiemann@uibk.ac.at> (2022-2023)
License: LGPL (see file COPYING.LESSER)
*)

theory Parallel_Closed_Impl
imports
  Parallel_Closed
  Critical_Pair_Closure_Impl
  Commutation
begin

fun is_par_rsteps_join where
  "is_par_rsteps_join R S (Some n) s t =
    (\<not> Option.is_none (List.find (\<lambda>v. (s,v) \<in> par_rstep (set R)) (reachable_terms S t n)))"
| "is_par_rsteps_join R S None s t =
    (\<not> Option.is_none (List.find (\<lambda>v. (s,v) \<in> par_rstep (set R)) (parallel_rewrite S t)))"

lemma is_par_rsteps_join[dest]: assumes "is_par_rsteps_join R S hint s t" 
  shows "\<exists>v. (s, v) \<in> par_rstep (set R) \<and> (t, v) \<in> (rstep (set S))\<^sup>*" 
proof (cases hint)
  case (Some n)
  let ?find = "\<lambda> v. (s,v) \<in> par_rstep (set R)" 
  from Some assms have "\<not> Option.is_none (List.find ?find (reachable_terms S t n))"  
    by auto
  then obtain v where "List.find ?find (reachable_terms S t n) = Some v"
    by force
  then have "(s,v) \<in> par_rstep (set R)" and "v \<in> set (reachable_terms S t n)"
    unfolding find_Some_iff by auto
  then show ?thesis
    using reachable_terms by blast
next
  case None
  let ?find = "\<lambda> v. (s,v) \<in> par_rstep (set R)" 
  from None assms have "\<not> Option.is_none (List.find ?find (parallel_rewrite S t))"
    by auto
  then obtain v where  "List.find ?find (parallel_rewrite S t) = Some v"
    by force
  then have "(s,v) \<in> par_rstep (set R)" and "v \<in> set (parallel_rewrite S t)"
    unfolding find_Some_iff by auto
  then show ?thesis using parallel_rewrite_par_step par_rstep_rsteps by blast
qed

declare is_par_rsteps_join.simps[simp del]

fun is_parallel_closed_cp where
  "is_parallel_closed_cp R _ (False, s, t) = ((s,t) \<in> par_rstep (set R))"
| "is_parallel_closed_cp R hint (True, s, t) = is_par_rsteps_join R R hint s t"  

context 
  fixes ren :: "'v :: {showl,Fresh.infinite} renaming2" 
begin
definition check_parallel_closed where
  "check_parallel_closed R n = do {
     check_left_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (is_parallel_closed_cp R n (b, s, t))
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not (almost) parallel closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl ren R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not (almost) parallel closed\<newline>'') \<circ> showsl_trs R)"

lemma check_parallel_closed:
  assumes "isOK(check_parallel_closed R n)"
  shows "CR (rstep (set R))"
proof (rule parallel_closed_imp_CR)
  from assms show "left_linear_trs (set R)" unfolding check_parallel_closed_def by auto
  show "\<And>p q. (False, p, q) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow> (p, q) \<in> par_rstep (set R)"
  proof -
    fix p q
    assume "(False, p, q) \<in> critical_pairs ren (set R) (set R)"
    with assms  show "(p, q) \<in> par_rstep (set R)" unfolding check_parallel_closed_def by auto
  qed
  show "\<And>p q. (True, p, q) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow>
           \<exists>v. (p, v) \<in> par_rstep (set R) \<and> (q, v) \<in> (rstep (set R))\<^sup>*"
  proof -
    fix p q
    assume *:"(True, p, q) \<in> critical_pairs ren (set R) (set R)"
    with assms[unfolded check_parallel_closed_def]
    have "is_parallel_closed_cp R n (True, p, q)" by auto 
    hence "is_par_rsteps_join R R n p q" by auto
    from is_par_rsteps_join[OF this]
    show "\<exists>v. (p, v) \<in> par_rstep (set R) \<and> (q, v) \<in> (rstep (set R))\<^sup>*" .
  qed
qed

definition check_parallel_closed_comm where
  "check_parallel_closed_comm R S n = do {
     check_left_linear_trs R;
     check_left_linear_trs S;
     check_allm (\<lambda> (b, s, t). do {
       check (is_par_rsteps_join S R n s t)
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' R<- . ->S '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not almost parallel closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl ren S R);
     check_allm (\<lambda> (b, s, t). do {
       check (b \<or> (s,t) \<in> par_rstep (set R))
         (showsl_lit (STR ''the inner critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' S<- . ->R '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not closed with parallel R step'') \<circ> showsl n)
     }) (critical_pairs_impl ren R S)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the almost parallel closed check for the following TRSs could not be proven\<newline>R: '') \<circ> 
        showsl_trs R o showsl_lit (STR ''\<newline>\<newline>S: '') o showsl_trs S)"

lemma check_parallel_closed_comm:
  assumes "isOK(check_parallel_closed_comm R S n)"
  shows "sig_commute F (set R) (set S)"
proof (intro commute_imp_sig_commute parallel_closed_commute is_par_rsteps_join[of _ _ n])
  note assms = assms[unfolded check_parallel_closed_comm_def, simplified]
  from assms show "left_linear_trs (set R)" "left_linear_trs (set S)" by auto
  from assms show "(b, p, q) \<in> critical_pairs ren (set S) (set R) \<Longrightarrow> is_par_rsteps_join S R n p q" for b p q
    by force
  from assms show "\<And>p q. (False, p, q) \<in> critical_pairs ren (set R) (set S) \<Longrightarrow> (p,q) \<in> par_rstep (set R)" 
    by force
qed

end
end